using System;
using System.Collections.Generic;
using System.Text;

namespace Oscar5Plugin
{
    class RateManager
    {
        private RateClass[] m_RateClasses;
        private Dictionary<ushort, RateClass> m_ClassLookup;
        private Dictionary<uint, RateClass> m_SNACLookup;

        private FLAPSocket m_Socket;

        public RateManager(FLAPSocket socket)
        {
            m_Socket = socket;
        }

        public byte[] GetRateClasses()
        {
            byte[] rateclasses = new byte[m_RateClasses.Length * 2];
            for (int i = 0; i < m_RateClasses.Length; i++)
            {
                Tools.PutWord(rateclasses, i * 2, m_RateClasses[i].ClassID);
            }
            return rateclasses;
        }


        //parses the reply to a rate parameters query
        public void ProcessRateParametersReply(FLAPPacket packet)
        {
            //grab the number of rate classes
            ushort classcount = Tools.GetWord(packet);

            //create the array and lookup tables for the classes
            m_RateClasses = new RateClass[classcount];
            m_ClassLookup = new Dictionary<ushort, RateClass>(classcount);
            m_SNACLookup = new Dictionary<uint, RateClass>();
            
            //read x rate parameter blocks
            for (int i = 0; i < classcount; i++)
            {
                //create a new rateclass object and read all the parameters into it
                RateClass r = new RateClass(m_Socket);
                ProcessRateParameters(r, packet);

                //add the rateclass to the array and the class lookup table
                m_RateClasses[i] = r;
                m_ClassLookup.Add(r.ClassID, r);
            }

            //read the associated snacs for each rate class
            for (int i = 0; i < classcount; i++)
            {
                //grab the class id the snacs are associated with and the number of snacs
                ushort classid = Tools.GetWord(packet);
                ushort snaccount = Tools.GetWord(packet);

                //look up the rateclass object by classid
                RateClass r = m_ClassLookup[classid];

                //read all the snacs associated with this class
                for (int x = 0; x < snaccount; x++)
                {
                    m_SNACLookup.Add(Tools.GetDword(packet), r);
                }
            }
        }

        //processes a rate change packet, this is sent by the server when we're approaching the rate limit,
        //or when we're rated
        public void ProcessRateChange(FLAPPacket packet)
        {
            //get the rate code, tells us why we're receiving this packet
            RateCode code = (RateCode)Tools.GetWord(packet);
            
            //look up the rateclass object by the classid
            RateClass rateclass = m_ClassLookup[Tools.GetWord(packet.Data, packet.Offset)];

            //load the new parameter info into the rateclass
            ProcessRateParameters(rateclass, packet);

            Console.WriteLine("ratecode: " + code.ToString());

            //now we decide what to do
            if (code == RateCode.Warning || code == RateCode.Limit || code == RateCode.Limit2)
            {
                rateclass.SetQueue();
            }
        }

        public void SendPacket(FLAPPacket packet)
        {
            RateClass r;

            //if there's no snac lookup table, or theres no rate class associated with that snac,
            //or there's no packet queue, send it along
            if (m_SNACLookup == null || !m_SNACLookup.TryGetValue(Tools.GetDword(packet.Data, 0), out r) || r.PacketTimerEnabled == false)
            {
                m_Socket.SendPacket(packet);
            }
            else //the rate class does have a queue, so add this packet to it
            {
                lock (r.PacketQueue) //acquire a lock on this queue since we're changing the data
                {
                    r.PacketQueue.Enqueue(packet);
                }
            }
        }

        //this reads the data from a rate parameters block into a rateclass object
        private void ProcessRateParameters(RateClass rateclass, FLAPPacket packet)
        {
            rateclass.ClassID = Tools.GetWord(packet);
            rateclass.WindowSize = Tools.GetDword(packet);
            rateclass.ClearLevel = Tools.GetDword(packet);
            rateclass.AlertLevel = Tools.GetDword(packet);
            rateclass.LimitLevel = Tools.GetDword(packet);
            rateclass.DisconnectLevel = Tools.GetDword(packet);
            rateclass.CurrentLevel = Tools.GetDword(packet);
            rateclass.MaxLevel = Tools.GetDword(packet);
            rateclass.LastTime = (uint)Environment.TickCount - Tools.GetDword(packet);
            rateclass.CurrentState = packet.Data[packet.Offset++];
        }
    }
}
